@@ -136,7 +136,7 @@ ospf_sk_open(struct ospf_iface *ifa)
   sk->flags = SKF_LADDR_RX | (ifa->check_ttl ? SKF_TTL_RX : 0);
   sk->ttl = ifa->cf->ttl_security ? 255 : 1;
 
-  if (sk_open(sk) < 0)
+  if (sk_open(sk, p->p.loop) < 0)
     goto err;
 
   /* 12 is an offset of the checksum in an OSPFv3 packet */
@@ -173,7 +173,7 @@ ospf_sk_open(struct ospf_iface *ifa)
 
  err:
   sk_log_error(sk, p->p.name);
-  rfree(sk);
+  sk_close(sk);
   return 0;
 }
 
@@ -220,7 +220,7 @@ ospf_open_vlink_sk(struct ospf_proto *p)
   sk->data = (void *) p;
   sk->flags = 0;
 
-  if (sk_open(sk) < 0)
+  if (sk_open(sk, p->p.loop) < 0)
     goto err;
 
   /* 12 is an offset of the checksum in an OSPFv3 packet */
@@ -234,7 +234,7 @@ ospf_open_vlink_sk(struct ospf_proto *p)
  err:
   sk_log_error(sk, p->p.name);
   log(L_ERR "%s: Cannot open virtual link socket", p->p.name);
-  rfree(sk);
+  sk_close(sk);
 }
 
 static void
@@ -311,7 +311,7 @@ ospf_iface_remove(struct ospf_iface *ifa)
 
   ospf_iface_sm(ifa, ISM_DOWN);
   rem_node(NODE ifa);
-  rfree(ifa->pool);
+  rp_free(ifa->pool);
 }
 
 void
@@ -484,9 +484,9 @@ ospf_iface_find(struct ospf_proto *p, struct iface *what)
 }
 
 static void
-ospf_iface_add(struct object_lock *lock)
+ospf_iface_add(void *_ifa)
 {
-  struct ospf_iface *ifa = lock->data;
+  struct ospf_iface *ifa = _ifa;
   struct ospf_proto *p = ifa->oa->po;
 
   /* Open socket if interface is not stub */
@@ -567,7 +567,7 @@ ospf_iface_new(struct ospf_area *oa, struct ifa *addr, struct ospf_iface_patt *i
     OSPF_TRACE(D_EVENTS, "Adding interface %s (%N) to area %R",
 	       iface->name, &addr->prefix, oa->areaid);
 
-  pool = rp_new(p->p.pool, "OSPF Interface");
+  pool = rp_new(p->p.pool, proto_domain(&p->p), "OSPF Interface");
   ifa = mb_allocz(pool, sizeof(struct ospf_iface));
   ifa->iface = iface;
   ifa->addr = addr;
@@ -668,8 +668,11 @@ ospf_iface_new(struct ospf_area *oa, struct ifa *addr, struct ospf_iface_patt *i
   lock->port = OSPF_PROTO;
   lock->inst = ifa->instance_id;
   lock->iface = iface;
-  lock->data = ifa;
-  lock->hook = ospf_iface_add;
+  lock->event = (event) {
+    .hook = ospf_iface_add,
+    .data = ifa,
+  };
+  lock->target = &global_event_list;
 
   olock_acquire(lock);
 }
@@ -687,7 +690,7 @@ ospf_iface_new_vlink(struct ospf_proto *p, struct ospf_iface_patt *ip)
 
   /* Vlink ifname is stored just after the ospf_iface structure */
 
-  pool = rp_new(p->p.pool, "OSPF Vlink");
+  pool = rp_new(p->p.pool, proto_domain(&p->p), "OSPF Vlink");
   ifa = mb_allocz(pool, sizeof(struct ospf_iface) + 16);
   ifa->oa = p->backbone;
   ifa->cf = ip;
@@ -1222,12 +1225,11 @@ ospf_ifa_notify3(struct proto *P, uint flags, struct ifa *a)
 static void
 ospf_reconfigure_ifaces2(struct ospf_proto *p)
 {
-  struct iface *iface;
   struct ifa *a;
 
-  WALK_LIST(iface, iface_list)
+  IFACE_WALK(iface)
   {
-    if (p->p.vrf_set && !if_in_vrf(iface, p->p.vrf))
+    if (p->p.vrf && !if_in_vrf(iface, p->p.vrf))
       continue;
 
     if (! (iface->flags & IF_UP))
@@ -1271,12 +1273,11 @@ ospf_reconfigure_ifaces2(struct ospf_proto *p)
 static void
 ospf_reconfigure_ifaces3(struct ospf_proto *p)
 {
-  struct iface *iface;
   struct ifa *a;
 
-  WALK_LIST(iface, iface_list)
+  IFACE_WALK(iface)
   {
-    if (p->p.vrf_set && !if_in_vrf(iface, p->p.vrf))
+    if (p->p.vrf && !if_in_vrf(iface, p->p.vrf))
       continue;
 
     if (! (iface->flags & IF_UP))
